#pragma TextEncoding="UTF-8"
#pragma rtGlobals=3
#pragma version=5.00
#pragma IgorVersion=8
#pragma ModuleName=tracer

// --------------------- Project Updater header ----------------------
// If you're using Igor Pro 8 or later and have Project Updater
// installed, this package can check periodically for new releases.
// https://www.wavemetrics.com/project/Updater
static constant kProjectID=342 // the project node on IgorExchange
static strconstant ksShortTitle="Tracer" // the project short title on IgorExchange

// https://www.wavemetrics.com/user/tony
// Please tell me when you find a bug, and let me know if you find this
// package useful or have suggestions for improvement.

// 5.00 Major rewrite! Send me your bug reports...

// Changed behaviour:

// Images are loaded to current data folder, not the package folder.
// Consequently Tracer can work with any images already present in
// current folder. The clear button still deletes only images loaded or
// created by Tracer.

// The International Commission on Illumination (CIE) ΔE*ab (1994)
// algorithm is used to quantify color difference.

// The interface for setting image scale has been tweaked. Right click to
// copy and paste image scaling.

// Command/ctrl-scroll to zoom. Option/Alt key accelerates zoom.

// The image plot is no longer constrained to display in plan mode.

// 4.20 A few minor changes, removed color-picker buttons because default
// color popup dialog has better selection tool with eye-dropper option.
// Added headers for Updater. The rest of this code could use a clean-up!
// 4.10 Uses /any flag with loadimage to load more image types
// 4.09 Bug fix: data folder not returned to starting one after quit on load.
// 4.08 Added headers for Project Updater.
// See http://www.igorexchange.com/project/Updater.
// 4.07 bug fix: tracing failed with error when starting with cursor on a
// non-trace pixel
// 4.06 Oct 9 2017 Smaller panel size + bug fix
// 4.05 Sep 28 2017 New option to force search for bounds of trace over
// some minimum range. Use this if you're trying to trace unfilled
// symbols.
// 4.04 Sep 22 2017 bug fix. allowing gaps could sometimes result in
// non-integer point numbers, leading to out of range errors.
// 4.03 6/23/17 create copies of image for editing
// 4.02 8/25/16 bug fix for editing RGBA images with an alpha
// (transparency) channel
// 4.01 6/20/16 fixed null string bug for empty display list at startup
// 4.00 4/12/16 Greyscale (2d) images worked in previous versions because
// Igor didn't complain about out of range errors, so layer 1 and 2
// values were the same as layer 0. Fixed my bad code and introduced some
// new features. Created a "pixel wiper" to assist with touching up image
// to help with trace extraction. Image is assumed to be 3D RGB or 2D
// greyscale wave. Now works with 24 or 48 bit images (8 or 16 bits per
// pixel). A future update for Igor 7/8/9 will use ScaleToIndex, maybe
// ImageInterpolate Warp...

// decrease kDoUpdateFrequency if you want to see the trace progression, 
// increase if speed is too slow
static constant kDoUpdateFrequency=10 
static constant kEditCopies=1
static constant kUsageHints=1

menu "Macros"
	"Tracer", /Q, tracer#Initialise()
end

// sets up package folder and creates globals
// then creates plot and panel
static function Initialise()
	NewDataFolder/O root:Packages
	NewDataFolder/O root:Packages:Tracer
	DFREF dfr = root:Packages:Tracer
	Make /O/N=11 dfr:sliderTicks=p/10
	Make /O/T/N=11 dfr:sliderLabels /wave=sliderLabels
	sliderLabels[0]="Exact"
	sliderLabels[10]="Fuzzy"
	DoWindow /K TracerGraph // start with a fresh window
	Display /K=1/N=TracerGraph /W=(200,10,550,350)  as "Tracer Image"
	makeTracerPanel()
end

static function makeTracerPanel()
	DFREF dfr = root:Packages:Tracer
	DoWindow /K TracerPanel
	NewPanel /K=2/N=TracerPanel/W=(190,0,0,450)/HOST=TracerGraph/EXT=1 as "Tracer Controls"
	ModifyPanel /W=TracerGraph#TracerPanel, noEdit=1
	
	variable v_l=15, v_t=5, groupw=182, font=12, deltaY=25, i=0
	
	GroupBox groupImages, win=TracerPanel,pos={v_l,v_t},size={groupw,80},title="Image selection", fsize=font
	GroupBox groupImages, win=TracerPanel,pos+={-10,0}
	i+=1
	Button Load, win=TracerPanel,pos={25,v_t+i*deltaY},size={60,20},Proc=tracer#tracerButtons,title="Load..."
	Button Load, win=TracerPanel,help={"Load an image to current data folder"}, fsize=font
	Button clean, win=TracerPanel,pos={105,v_t+i*deltaY},size={60,20},title="Clear...", Proc=tracer#tracerButtons
	Button clean, win=TracerPanel,help={"Kill all images created by Tracer"}, fsize=font
	i+=1
	PopupMenu popImage, win=TracerPanel, mode=1, Value=tracer#getImgList(), pos={v_l,v_t+i*deltaY},size={190,20}
	PopupMenu popImage, win=TracerPanel, Proc=tracer#tracerPopMenu, title="Display: ", fsize=font
	PopupMenu popImage, win=TracerPanel, help={"Display an image from current data folder"}
	i+=1.5
		
	GroupBox groupScale, win=TracerPanel,pos={v_l,v_t},size={groupw,45},title="Set Scale", fsize=font
	GroupBox groupScale, win=TracerPanel,pos+={-10,i*deltaY}
	i+=1
	Button startScale, win=TracerPanel,pos={20,i*deltaY},size={30,20},Proc=tracer#tracerButtons,title="Off",fColor=(65535,65533,65534)
	Button startScale, win=TracerPanel, help={"Click to toggle setscale cursors"}
	
	CheckBox checkLogX, win=TracerPanel,pos={60,i*deltaY},size={36,14},Proc=tracer#tracerCheckBoxes,title="Log X"
	CheckBox checkLogX, win=TracerPanel,value=0, fsize=font, help={"Check for logarithmic X axis"}
	CheckBox checkLogY, win=TracerPanel,pos={115,i*deltaY},size={36,14},Proc=tracer#tracerCheckBoxes,title="Log Y"
	CheckBox checkLogY, win=TracerPanel,value=0, fsize=font, help={"Check for logarithmic Y axis"}
	
	i+=1
	GroupBox groupColor, win=TracerPanel,pos={v_l,v_t},size={groupw,95},title="Colors", fsize=font
	GroupBox groupColor, win=TracerPanel,pos+={-10,i*deltaY}
	i+=1
	PopupMenu popTraceRGB, win=TracerPanel,pos={54,i*deltaY},size={50,21},Proc=tracer#tracerPopMenu, title="data", bodywidth=50
	PopupMenu popTraceRGB, win=TracerPanel,mode=1,popColor= (0,0,0),value= #"\"*COLORPOP*\"", fsize=font
	PopupMenu popTraceRGB, win=TracerPanel,help={"select \"Other...\", then the eye dropper to select color with mouse"}
	PopupMenu popBGRGB, win=TracerPanel,pos={125,i*deltaY},size={50,21},Proc=tracer#tracerPopMenu, title="bg", bodywidth=50
	PopupMenu popBGRGB, win=TracerPanel,mode=1,popColor=(65535,65535,65535),value= #"\"*COLORPOP*\"", fsize=font
	PopupMenu popBGRGB, win=TracerPanel,help={"select other, then the eye dropper to select color with mouse"}
	i+=1
	Slider sliderFuzzy, win=TracerPanel,pos={18,i*deltaY},size={153,42},fSize=font-2
	Slider sliderFuzzy, win=TracerPanel,limits={0,1,0.05},vert= 0
	Slider sliderFuzzy, win=TracerPanel,userTicks={dfr:sliderTicks,dfr:sliderLabels}
	Slider sliderFuzzy, win=TracerPanel,value=0.3, help={"increase fuzziness to accept a greater range of pixel colors"}
	
	i+=2
	GroupBox groupExtract, win=TracerPanel,pos={v_l,v_t},size={groupw,150},title="Extract", fsize=font
	GroupBox groupExtract, win=TracerPanel,pos+={-10,i*deltaY}
	i+=1
	SetVariable setvarTraceName, win=TracerPanel,pos={18,i*deltaY},size={155,16},Proc=tracer#tracerSetVar,title="Create:"
	SetVariable setvarTraceName, win=TracerPanel,value=_STR:UniqueName("trace",1,0), fsize=font, help={"set name for new trace"}
	i+=1
	CheckBox checkXY, win=TracerPanel,pos={v_l,i*deltaY},size={58,14},Proc=tracer#tracerCheckBoxes,title="XY Data"
	CheckBox checkXY, win=TracerPanel,value=0, help={"create X and Y waves"}, fsize=font
	CheckBox checkGaps, win=TracerPanel,pos={90,i*deltaY},size={71,14},Proc=tracer#tracerCheckBoxes,title="Allow Gaps", fsize=font
	CheckBox checkGaps, win=TracerPanel,value=0, help={"allow tracing to continue when no target pixel is found"}
	i+=1
	SetVariable setvarJump, win=TracerPanel,pos={25,i*deltaY},size={140,16},title="Jump threshold"
	SetVariable setvarJump, win=TracerPanel,limits={0,Inf,0},value=_NUM:100, fsize=font
	SetVariable setvarJump, win=TracerPanel,help={"trace-colored pixels must be found within this range of\rprevious pixel to be considered part of continuous trace"}
	i+=1
	SetVariable setvarRange, win=TracerPanel,pos={25,i*deltaY},size={140,16},title="Minimum range"
	SetVariable setvarRange, win=TracerPanel,limits={0,Inf,0},value=_NUM:0, fsize=font
	SetVariable setvarRange, win=TracerPanel,help={"Force search to consider all pixels within range when\rsearching for bounds of trace (default = 0)"}
	i+=1
	Button Go, win=TracerPanel,pos={64,i*deltaY},size={50,20},Proc=tracer#tracerButtons,title="Trace!", fsize=font, fColor=(57346,65535,49151), help={"start tracing from cursor location"}

//	editing controls
	i+=1
	GroupBox groupEdit, win=TracerPanel,pos={v_l-10,v_t+i*deltaY},size={groupw,50},title="Edit", fsize=font
	i+=1.1
	Button startEditImg, win=TracerPanel,pos={20,i*deltaY},size={70,20},Proc=tracer#tracerButtons,title="Edit image", fsize=font
	Button startEditImg, win=TracerPanel,help={"Edit the image to remove obstacles to tracing"}
	Button startEditTrace, win=TracerPanel,pos={100,i*deltaY},size={70,20},Proc=tracer#tracerButtons,title="Edit trace", fsize=font
	Button startEditTrace, win=TracerPanel,help={"Adjust positions of points in traced curve after tracing"}
	
	// if we have already loaded an image, plot it
	ControlInfo /W=TracerGraph#TracerPanel popImage
	wave /Z w_img=$S_Value
	if (WaveExists(w_img))
		PlotImage(w_img)
	endif
end

static function fillPanelStructure(STRUCT PanelStatusStructure &s)
	getControlValue("TracerGraph#TracerPanel", "popImage", s.image)
	getControlValue("TracerGraph#TracerPanel", "checkLogX", s.logX)
	getControlValue("TracerGraph#TracerPanel", "checkLogY", s.logY)
	getControlValue("TracerGraph#TracerPanel", "popTraceRGB", s.trace)
	getControlValue("TracerGraph#TracerPanel", "popBGRGB", s.bg)
	getControlValue("TracerGraph#TracerPanel", "sliderFuzzy", s.fuzzy)
	getControlValue("TracerGraph#TracerPanel", "checkGaps", s.gaps)
	getControlValue("TracerGraph#TracerPanel", "checkXY", s.XY)
	getControlValue("TracerGraph#TracerPanel", "setvarJump", s.jump)
	getControlValue("TracerGraph#TracerPanel", "setvarRange", s.range)
	getControlValue("TracerGraph#TracerPanel", "setvarTraceName", s.tracename)
end

static structure PanelStatusStructure
	STRUCT ControlValueStructure image, logX, logY, trace, bg, fuzzy, gaps, XY, jump, range, tracename
endstructure

static structure ControlValueStructure
	int16 type
	variable value
	string sval
	STRUCT RGBColor rgb
	variable selcol
endstructure

static function getControlValue(string win, string controlName, STRUCT ControlValueStructure &s)
	ControlInfo /W=$win $controlName
	switch (abs(V_Flag))
		case 3: // popup menu
			s.rgb.red = V_Red
			s.rgb.green = V_Green
			s.rgb.blue = V_Blue
//			s.rgb.alpha = V_Alpha
		case 2: // checkbox
		case 4: // valdisplay
		case 5: // SetVariable
		case 7: // slider
		case 8: // tab
			s.type = v_flag
			s.value = v_value
			s.sval = s_value
			break
		case 11: // listbox
			s.type = v_flag
			s.value = v_value
			s.sval = s_value
			s.selcol = v_selcol
			break
		default:
			s.type = v_flag
	endswitch
end

// get list of images in current datafolder
static function /S getImgList([int tracer])
	tracer = ParamIsDefault(tracer) ? 0 : tracer
	string ListOfImages = WaveList("*", ";", "DIMS:2") + WaveList("*", ";", "DIMS:3")
	if (tracer)
		int i
		for (i=ItemsInList(ListOfImages)-1;i>=0;i--)
			wave w = $StringFromList(i, ListOfImages)
			if (NumberByKey("TracerImage", note(w), "=") !=1 )
				ListOfImages = RemoveListItem(i, ListOfImages)
			endif
		endfor
		return ListOfImages
	endif
	if (strlen(ListOfImages) == 0)
		ListOfImages = "none;"
	endif
	return ListOfImages
end

static function tracerButtons(STRUCT WMButtonAction &s)
	if (s.eventcode == -1)
		KillDataFolder /Z root:packages:tracer:
		return 0
	endif
		
	if (s.eventCode!=2)
		return 0
	endif
			
	strswitch(s.ctrlName)
		case "load":
			LoadTracerImage()
			return 0
		case "clean":
			Cleanup()
			return 0
	endswitch
	
	if (strlen(ImageNameList("TracerGraph", ";" ))==0)
		DoAlert 0, "No image display!"
		return 0
	endif
	
	DFREF dfr = root:Packages:Tracer
	string cmd, str
	
	strswitch(s.ctrlName)
		case "Go":
			ExtractTrace()
			break
		case "startEditTrace":
			EditTrace(1)
			break
		case "stopEditTrace":
			EditTrace(0)
			break
		case "startEditImg":
			EditImage(1)
			break
		case "stopEditImg":
			EditImage(0)
			break
		case "startScale":
			SetImgScale(1)
			break
		case "stopScale":
			SetImgScale(0)
			break
	endswitch
	return 0
end

static function EditImage(int start)
	SetWindow TracerGraph hook(EditImgHook)=$""
	ControlInfo /W=TracerPanel StopEditImg
	if(v_flag == 1)
		Button StopEditImg, win=TracerGraph#TracerPanel, fColor=(0,0,0), title="Edit image", Rename=startEditImg
	endif
	if (start == 0)
		return 0
	endif
	
	ControlInfo /W=TracerGraph#TracerPanel popImage
	string strImage = S_Value
	wave /Z w_img=$strImage
	if (WaveExists(w_img)==0)
		return 0
	endif
	
	int MakeCopy=kEditCopies, reedit=0

	if (MakeCopy && cmpstr(strImage, GetUserData("TracerGraph","","editing"))==0)
		DoAlert 2, "continue editing " + strImage + "?\r\rNo will create a copy"
		if (V_flag == 3)
			return 0
		elseif(V_flag == 1)
			MakeCopy = 0 // don't make a copy this time
		endif
		reedit = 1
	endif
		
	if (kUsageHints && reedit==0)
		string cmd = "Warning: you are about to edit " + SelectString(MakeCopy, "", "a copy of ") + strImage
		cmd+="\r\rHold ctrl/option and move mouse to add trace/bg colour. Expand view for detailed editing.\r\r"
		cmd+="Continue?"
		DoAlert 1, cmd
		if (v_flag==2)
			return 0
		endif
	endif
	
	if(MakeCopy)
		string newName = UniqueName(NameOfWave(w_img),1,0)
		Duplicate w_img $newName /wave=w
		note /K w ReplaceNumberByKey("TracerImage",note(w),1,"=")
		
		// save current axis limits
		variable Lmin, Lmax, Bmin, Bmax
		GetAxis /W=TracerGraph/Q left
		Lmin=V_min;Lmax=V_max
		GetAxis /W=TracerGraph/Q bottom
		Bmin=V_min;Bmax=V_max
		
		PopupMenu popImage, win=TracerGraph#TracerPanel, popmatch=NameOfWave(w)
		PlotImage(w)
		
		// restore axis limits
		SetAxis /W=TracerGraph left, Lmin, Lmax
		SetAxis /W=TracerGraph bottom, Bmin, Bmax
		strImage=newName
	endif
	
	DoWindow /F TracerGraph
	SetWindow TracerGraph hook(EditImgHook)=tracer#hookEditImg, userdata(editing)=strImage
	Button startEditImg, win=TracerGraph#TracerPanel, fcolor=(65535,0,0), title="Stop edit", Rename=StopEditImg
	
	return 1
end

static function LoadTracerImage()
	DFREF dfr = root:Packages:Tracer
	ImageLoad	/Q/T=any
	if (V_Flag==0)
		return 0
	endif
	string imageStr=StringFromList(0, S_waveNames)
	string newImageStr=ParseFilePath(3, imageStr, ":", 0, 0)
	newImageStr=CleanupName(newImageStr,0)
	if (CheckName(newImageStr, 1))
		newImageStr=UniqueName(newImageStr, 1, 0)
	endif
	Rename $imageStr $newImageStr
	wave w_img=$newImageStr
	note w_img "TracerImage=1;"
	PlotImage(w_img)
	return 1
end

static function PlotImage(wave w_img)
	STRUCT PanelStatusStructure s
	fillPanelStructure(s)
	SetImgScale(0)
	EditImage(0)
	EditTrace(0)
	ClearPlot()
	
	AppendImage /W=TracerGraph w_img
	string str=NameOfWave(w_img)
	
	SetAxis /A
	if (DimDelta(w_img, 0) < 0)
		SetAxis/A/R bottom
	endif
	if (DimDelta(w_img, 1) > 0)
		SetAxis/A/R left
	endif
	
	ModifyGraph width={Plan, abs(DimDelta(w_img, 1 )/ DimDelta(w_img, 0)) ,bottom,left}
	DoUpdate
	ModifyGraph width=0
	
	if(s.LogX.value || s.LogY.value)
		RebuildGraph()
	endif

	Cursor /N=1/I/W=TracerGraph/C=(65535,0,0)/P A $str 0.15*DimSize(w_img, 0), 0.5*DimSize(w_img,1)
	Cursor /N=1/I/W=TracerGraph/C=(0,65535,0)/P B $str 0.85*DimSize(w_img, 0), 0.5*DimSize(w_img,1)

	SetWindow TracerGraph hook(tracerHook)=tracer#hookTracer
	
	PopupMenu popImage, win=TracerGraph#TracerPanel, Value=tracer#getImgList()
	PopupMenu popImage, win=TracerGraph#TracerPanel, popmatch=NameOfWave(w_img)
	return 1
end

static function ClearPlot()
	string strImage="", strTrace=""
	do
		strImage=StringFromList (0, ImageNameList("TracerGraph", ";" ))
		RemoveImage /Z/W=TracerGraph $strImage
	while(strlen(strImage))
	do
		strTrace=StringFromList (0, TraceNameList("TracerGraph", ";" ,1))
		RemoveFromGraph /Z/W=TracerGraph $strTrace
	while(strlen(strTrace))
	return 1
end

static function Cleanup()
	string imageList = getImgList(tracer=1)
	if(ItemsInList(imageList)==0)
		return 0
	endif
	string cmd="The following image waves will be deleted:"
	string tooMany = "\r*** TOO MANY TO LIST ***"
	string cont = "\r\rContinue?"
	string imageName
	int numImages = ItemsInList(imageList)
	int cmdLen = strlen(cmd) + strlen(tooMany) + strlen(cont)
	int i
	for (i=0;i<numImages;i+=1)
		imageName = StringFromList(i, imageList)
		// avoid too-long alert
		if ( cmdLen+strlen(imageName) >= 1023 )
			cmd += tooMany
			break
		endif
		cmd += "\r" + imageName
		cmdLen += strlen("\r" + imageName)
	endfor
	cmd += cont
	
	DoAlert 1, cmd
	if (V_Flag==2)
		return 0
	endif
	ClearPlot()
	for (i=0;i<ItemsInList(imageList); i+=1)
		KillWaves /Z $StringFromList(i, imageList)
	endfor
	SetWindow TracerGraph userdata(imageList) = ""
	ControlUpdate /W=TracerGraph#TracerPanel popImage
	return 1
end

static function tracerCheckBoxes(STRUCT WMCheckboxAction &s)
	if (s.eventCode!=2)
		return 0
	endif
	strswitch (s.ctrlName)
		case "CheckLogX":
			if (s.checked)
				CheckBox checkXY, win=TracerGraph#TracerPanel, value=1, disable=2
			else
				CheckBox checkXY, win=TracerGraph#TracerPanel, value=0, disable=0
			endif
		case "CheckLogY": // logX or logY
			RebuildGraph()
	endswitch
	return 0
end

// rebuild graph without changing image
// allows switch between log and normal axes
static function RebuildGraph()
	DoWindow tracergraph
	if (V_Flag==0)
		return 0
	endif
	
	STRUCT PanelStatusStructure s
	fillPanelStructure(s)
	
	DFREF dfr = root:Packages:Tracer
	string strImage=StringFromList (0, ImageNameList("TracerGraph", ";"))
	wave /Z w_image=ImageNameToWaveRef("TracerGraph", strImage)
	if (WaveExists(w_image)==0)
		return 0
	endif
	
// save cursor positions and make sure they end up on the same pixels
	Make /free/n=6 csrp=NaN, csrq=NaN
	Make /free/T csr ={"A","B","C","D","E","F"}
	int i
	for(i=0;i<6;i++)
		if (cmpstr(StringByKey("TNAME", CsrInfo($csr[i])), strImage) == 0) // cursor is on image
			csrp[i]=pcsr($csr[i])
			csrq[i]=qcsr($csr[i])
		endif
	endfor
		
	RemoveImage /W=tracergraph $strImage
	string ImageNote=note(W_Image)
	variable Xhigh, Xlow, Yhigh, Ylow
	
	if (s.LogX.value)
		Make /O/N=(DimSize(w_image,0)+1) dfr:TracerLogX /wave=TracerLogX
		Xhigh=NumberByKey("Xhigh", ImageNote)
		if (numtype(Xhigh)==2)  //nan
			Xhigh=100
		endif
		Xlow=NumberByKey("Xlow", ImageNote)
		if (numtype(Xlow)==2)  //nan
			Xlow=1
		endif
		TracerLogX=alog(log(Xlow)+(p+0.5)/DimSize(TracerLogX,0)*(log(Xhigh)-log(Xlow)))
	endif
	if (s.LogY.value)
		Make /O/N=(DimSize(w_image,1)+1) dfr:TracerLogY /wave=TracerLogY
		Yhigh=NumberByKey("Yhigh", ImageNote)
		if (numtype(Yhigh)==2)  //nan
			Yhigh=1
		endif
		Ylow=NumberByKey("Ylow", ImageNote)
		if (numtype(Ylow)==2)  //nan
			Ylow=100
		endif
		TracerLogY=alog(log(Ylow)+(DimSize(TracerLogY,0)-p-0.5)/DimSize(TracerLogY,0)*(log(Yhigh)-log(Ylow)))
	endif
	
	// need to specify this for Igor 7
	ModifyGraph  /W=TracerGraph width=0
	
	AppendImage /W=TracerGraph w_image vs {dfr:$SelectString(s.LogX.value, "*", "TracerLogX"),dfr:$SelectString(s.LogY.value, "*", "TracerLogY")}
	if ((s.LogX.value && TracerLogX[0]>TracerLogX[1]) || (s.LogX.value==0 && DimDelta(w_image, 0) < 0))
		SetAxis/A/R bottom
	endif
	if ((s.LogY.value && TracerLogY[1]>TracerLogY[0]) || (s.LogY.value==0 && DimDelta(w_image, 1) > 0))
		SetAxis/A/R left
	endif
	ModifyGraph /W=TracerGraph log(left)=s.LogY.value, log(bottom)=s.LogX.value
		
	if (s.LogX.value || s.LogY.value)
		ModifyGraph /W=TracerGraph width=0
	else
		ModifyGraph /W=TracerGraph width={Plan, abs(DimDelta(w_image, 1 )/ DimDelta(w_image, 0)) ,bottom,left}
		DoUpdate
		ModifyGraph /W=TracerGraph width=0
	endif
	
	for(i=0;i<6;i++)
		if (numtype(csrp[i]) == 0) // cursor was on image
			Cursor /I/W=TracerGraph/P $csr[i] $strImage csrp[i], csrq[i]
		endif
	endfor
	DoUpdate
	if (numtype(csrp[2]) == 0) // set scale is active
		STRUCT WMWinHookStruct swin
		swin.WinName="TracerGraph"
		swin.eventcode=6
		hookSetScale(swin)
	endif
	return 1
end

static function tracerSetVar(STRUCT WMSetVariableAction &s)
	if(s.eventCode!=8)
		return 0
	endif
	SetVariable $s.ctrlName, win=$s.win, value=_STR:CleanupName(s.sval, 0)
	return 1
end

static function EditTrace(int start)
	GraphNormal /W=TracerGraph
	ControlInfo /W=TracerPanel StopEditTrace
	if(v_flag == 1)
		Button StopEditTrace, win=TracerGraph#TracerPanel, fColor=(0,0,0), title="Edit trace", Rename=startEditTrace
	endif
	if (start == 0)
		return 0
	endif
	string traceName=""
	if (!(ItemsInList( TraceNameList("TracerGraph",";",1))))
		DoAlert 0, "No traces on graph"
		return 0
	endif
	Prompt traceName, "Trace to edit:"Popup, TraceNameList("TracerGraph",";",1)
	DoPrompt "Select trace to edit", traceName
	if (V_flag)
		return 0
	endif
	GraphWaveEdit /W=TracerGraph /M $traceName
	Button startEditTrace, win=TracerGraph#TracerPanel, fcolor=(65535,0,0), title="Stop edit", Rename=StopEditTrace
	return 1
end

static function tracerPopMenu(STRUCT WMPopupAction &s)
	if (s.eventCode!=2)
		return 0
	endif
	if (stringmatch(s.ctrlName, "popImage"))
		CheckBox checkLogX win=tracerPanel, value=0
		CheckBox checkLogY win=tracerPanel, value=0
		CheckBox checkXY win=tracerPanel, value=0, disable=0
		wave /Z w_img=$s.popStr
		if (WaveExists(w_img))
			PlotImage(w_img)
		endif
		return 0
	endif
	
	// colour pickers
	ControlInfo /W=TracerGraph#TracerPanel $s.ctrlName
	wave /Z w_img=GetImageRef()
	if (WaveExists(w_img) && (DimSize(w_img,2)==0))
		variable v_grey=(V_Red+V_Green+V_Blue)/3
		PopupMenu $s.ctrlName, win=$s.win, popColor=(v_grey,v_grey,v_grey)
	endif
	return 0
end

// this is where the magic happens
static function ExtractTrace()
	
	STRUCT PanelStatusStructure s
	fillPanelStructure(s)
	
	DFREF dfr = root:Packages:Tracer
	string ImageNameString=StringFromList (0, ImageNameList("TracerGraph", ";" ))
	wave /Z w_img=ImageNameToWaveRef("TracerGraph",ImageNameString)

	//ShowInfo /W=TracerGraph
	if ( strlen(CsrInfo(A, "TracerGraph"))==0 || strlen(CsrInfo(B, "TracerGraph"))==0 )
		DoAlert 0, "Set cursors at start and end of trace"
		return 0
	endif
	if (cmpstr(CsrWave(A, "TracerGraph"), CsrWave(B, "TracerGraph")))
		DoAlert 0, "Cursors must be on the same image"
		return 0
	endif
	if (pcsr(A, "TracerGraph")>pcsr(B, "TracerGraph"))
		variable tempP=pcsr(A, "TracerGraph"), tempQ=qcsr(A, "TracerGraph")
		Cursor /I /P /W=TracerGraph A $CsrWave(A, "TracerGraph") pcsr(B, "TracerGraph"), qcsr(B, "TracerGraph")
		Cursor /I /P /W=TracerGraph B $CsrWave(A, "TracerGraph") tempP, tempQ
	endif
	
	variable resolution=abs(pcsr(B, "TracerGraph")-pcsr(A, "TracerGraph"))+1
	
	if (CheckName(s.tracename.sval, 1) )
		DoAlert 1, s.tracename.sval + " already exists. Overwrite?"
		if (V_Flag==2)
			return 0
		endif
	endif
	
	Make /O/N=(resolution) $s.tracename.sval=NaN
	wave dataWave=$s.tracename.sval
	Print "created wave "+NameOfWave(dataWave)
	
	RemoveFromGraph /Z /W=TracerGraph $NameOfWave(dataWave) // just in case
	
	// *********
	// should stick with point scaling for XY data
	
	if (!s.logX.value)
		SetScale /I x, xcsr(A, "TracerGraph"), xcsr(B, "TracerGraph"), dataWave
	endif
	
	if (s.XY.value)
		Make /O/N=(resolution) $s.tracename.sval+"_X"=NaN
		wave dataXwave=$s.tracename.sval+"_X"
		SetScale /I x, xcsr(A, "TracerGraph"), xcsr(B, "TracerGraph"), dataXwave
		
		if (s.logX.value) // deal with X data for log X plots
			wave TracerLogX=dfr:TracerLogX
			dataXwave=TracerLogX[pcsr(A)+p+0.5]
		else
			dataXwave=x
		endif
	endif
		
	string info=ImageInfo("TracerGraph", ImageNameString, 0)
	
	if(s.XY.value)
		AppendToGraph /W=TracerGraph dataWave vs dataXwave
	else
		AppendToGraph /W=TracerGraph dataWave
	endif
	
	STRUCT RGBcolor RGB
	RGB=s.trace.rgb
	contrastingColor(RGB)
	ModifyGraph /W=TracerGraph rgb($NameOfWave(dataWave))=(rgb.red,rgb.green,rgb.blue)
	ModifyGraph /W=TracerGraph mode($NameOfWave(dataWave))=0,lsize($NameOfWave(dataWave))=2
	
	
	variable i=pcsr(A, "TracerGraph")
	variable j=0 // DataWave point number

	variable highP=qcsr(A, "TracerGraph"), lowP=qcsr(A, "TracerGraph")
	
	variable gap, jump=0, k, p_Y
	
	do
		gap=0
		if ((RGBgood(w_img, i, highP, s)==0)) // non-trace pixel
			do
				// expand to find a trace pixel
				lowP-=1
				highP+=1
							
				if (RGBgood(w_img, i, highP, s))
					lowP=highP
					break
				elseif(RGBgood(w_img, i, lowP, s))
					highP=lowP
					break
				endif
			while ( highP<(DimSize(w_img,1)-1) && lowP>1 )
		endif
		// check that we're on a trace pixel now
			
		if (highP>lowP) // failed to find one
			if (s.gaps.value)
				gap=1
			else
				if (s.XY.value)
					DoAlert 0, "Failed to follow trace at x=" +  num2str(dataXwave[j])+ "\rTry allowing gaps or adjusting fuzziness setting"
				else
					DoAlert 0, "Failed to follow trace at x=" +  num2str(pnt2x(dataWave, j))+ "\rTry allowing gaps or adjusting fuzziness setting"
				endif
				return 0
			endif
		else // found a trace pixel
			
			k=max(j,1)
			do
				k-=1 // find the last non-nan value in dataWave
			while((numtype(dataWave[k])==2)&&k)
			
			if (s.logY.value)
				wave TracerLogY=dfr:TracerLogY
				if (numtype(dataWave[k])==0)
					p_Y=BinarySearchInterp(TracerLogY, dataWave[k])
					// p_Y should be image column number

					if(abs(highP-p_Y)>s.jump.value)
						// more than JumpThreshold pixels away from previous value
						if (s.gaps.value)
							gap=1 // treat it as a gap
						else
							if (s.logX.value)
								wave TracerLogX=dfr:TracerLogX
								jump+=(jump==0)*TracerLogX[pcsr(A)+j+0.5]
							else
								jump+=(jump==0)*pnt2x(dataWave, j) // save first jump position
							endif
						endif
					endif
				endif
			elseif (abs(highP-(dataWave[k]-DimOffset(w_img, 1))/DimDelta(w_img, 1))>s.jump.value )
				// more than JumpThreshold pixels away from previous value
				if (s.gaps.value)
					gap=1 // treat it as a gap
				else
					if (s.logX.value)
						wave TracerLogX=dfr:TracerLogX
						jump+=(jump==0)*TracerLogX[pcsr(A)+j+0.5]
					else
						jump+=(jump==0)*pnt2x(dataWave, j) // save first jump position
					endif
				endif
			endif
		endif
		if (gap)     // j is datawave point number, i is img x pixel
			k=max(j,1)
			do
				k-=1 // find the last non-nan value in dataWave
			while(numtype(dataWave[k])==2 && k)
			
			dataWave[j]=NaN
			
			if (numtype(datawave[k])) // must have failed to find first point
				highP=qcsr(A, "TracerGraph") // keep heading right until we hit the trace
			else
				if (s.logX.value)
					wave TracerLogY=dfr:TracerLogY
					highP=BinarySearch(TracerLogY, dataWave[k])
				else
					highP=round((dataWave[k]-DimOffset(w_img, 1))/DimDelta(w_img, 1))
				endif
			endif
			lowP=highP
		else		// not a gap
			if (s.range.value)
				highP+=floor(s.range.value/2)
				lowP-=floor(s.range.value/2)
			endif	
			highP = min(DimSize(w_img,1)-2, highP)
			lowP = max(1, lowP)
			do // expand to find vertical limits of trace
				highP+=RGBgood(w_img, i, highP+1, s) // increment only if next pixel is good
				lowP-=RGBgood(w_img, i, lowP-1, s)
							
				lowP=max(lowP,0)
				highP=min(highP,DimSize(w_img,1)-1)
				
				if ( RGBgood(w_img,i,highP+1, s)==0 && RGBgood(w_img,i,lowP-1, s)==0 )
					// if we started on a trace pixel,
					// we've found the upper and lower bounds of trace
					break
				endif
			while (highP<(DimSize(w_img,1)-1) && lowP>0) // stop if we reach edge of image
			if (s.range.value) // contract inward to find limits of trace
				do
					highP-=(RGBgood(w_img,i,highP, s)==0)
					lowP+=(RGBgood(w_img,i,lowP, s)==0)
				while( (RGBgood(w_img,i,highP, s)==0) || (RGBgood(w_img,i,lowP, s)==0) )
			endif

			// find the mid-point
			if (s.logY.value)
				wave TracerLogY=dfr:TracerLogY
				dataWave[j]=TracerLogY[(highP+LowP)/2+0.5]
			else
				dataWave[j]=DimOffset(w_img, 1)+DimDelta(w_img, 1)*(highP+LowP)/2
			endif
			
			highP=round((highP+LowP)/2)
			lowP=highP
		endif
		// allowing the doupdate can slow things down
		if((kDoUpdateFrequency) && mod(i, kDoUpdateFrequency)==0) // sparse updating
			DoUpdate
		endif
		i+=1
		j+=1
	while(i<=pcsr(B, "TracerGraph"))
	
	// make sure we found something!
	variable badData=1
	for (j=0;j<DimSize(dataWave, 0);j+=1)
		if(numtype(dataWave[j])==0)
			badData=0
			break
		endif
	endfor
	
	if (jump!=0)
		Print "Failed to follow trace at x=" +  num2str(jump)
		DoAlert 1, "May have failed to follow trace at x=" +  num2str(jump)+ "\rTry allowing gaps or adjusting jump threshold\rDelete "+ NameOfWave(dataWave)+"?"
		if (V_flag==1)
			RemoveFromGraph /W=TracerGraph $NameOfWave(dataWave)
			KillWaves dataWave
			Print "deleted wave " + s.tracename.sval
		endif
	elseif ( badData || (abs(qcsr(B)- highP )>10) ) // no data, or more than 10 pixels away from target
		Print "Looks like trace failed"
		DoAlert  1, "Looks like trace failed.\rDelete " + NameOfWave(dataWave) + "?"
		if (V_flag==1)
			RemoveFromGraph /W=TracerGraph $NameOfWave(dataWave)
			KillWaves dataWave
			Print "deleted wave " + s.tracename.sval
		endif
	else
		Print "Trace extracted successfully"
	endif
end

// determine whether w_img[pixelX][pixelY] is close enough in color to be considered a trace pixel
static function RGBgood(wave w_img, variable pixelX, variable pixelY, STRUCT PanelStatusStructure &s)
	STRUCT RGBcolor RGBpixel
	getPixelRGB(RGBpixel, w_img, pixelX, pixelY)
	STRUCT LABcolor LABdata
	STRUCT LABcolor LABpixel
	RGB2LAB(RGBpixel, LABpixel)
	RGB2LAB(s.trace.rgb, LABdata)
	return deltaE94(LABdata, LABpixel) < (s.fuzzy.value*50)
end

static function getPixelRGB(STRUCT RGBcolor &RGB, wave w_img, variable pixelX, variable pixelY)
	// work with 24 or 48 bit images (8 or 16 bits per pixel)
	variable bitdepth = (WaveType(w_img)&8) ? 257 : 1
	if (DimSize(w_img,2))
		RGB.red=w_img[pixelX][pixelY][0]*bitdepth
		RGB.green=w_img[pixelX][pixelY][1]*bitdepth
		RGB.blue=w_img[pixelX][pixelY][2]*bitdepth
	else
		int grey=w_img[pixelX][pixelY]*bitdepth
		RGB.red=grey
		RGB.green=grey
		RGB.blue=grey
	endif
	return 1
end

// write RGB values to a pixel, or a bunch of pixels,
// depending on how zoomed-in we are on the image
static function writePixelRGB(STRUCT RGBcolor &RGB, wave w_img, variable pixelX, variable pixelY, STRUCT PanelStatusStructure &s)
	DFREF dfr = root:Packages:Tracer
	
	// figure out colour depth
	// work with 24 or 48 bit images (8 or 16 bits per pixel)
	Make /free w={RGB.red,RGB.green,RGB.blue}
	w /= WaveType(w_img)&8 ? 257 : 1
	
	// figure out zoom level
	variable hpoints, vpoints
	GetAxis /W=TracerGraph /Q bottom
	if (s.logX.value)
		wave TracerLogX=dfr:TracerLogX
		hpoints = abs(BinarySearch(TracerLogX, v_max)-BinarySearch(TracerLogX, v_min))
	else
		hpoints = abs(scaleToIndex(w_img, v_max, 0)-scaleToIndex(w_img, v_min, 0))
	endif
	GetAxis /W=TracerGraph /Q left
	if (s.logY.value)
		wave TracerLogY=dfr:TracerLogY
		vpoints = abs(BinarySearch(TracerLogY, v_max)-BinarySearch(TracerLogY, v_min))
	else
		vpoints = abs(scaleToIndex(w_img, v_max, 1)-scaleToIndex(w_img, v_min, 1))
	endif
	GetWindow tracergraph gsize
	variable zoom=hpoints/(v_right-v_left) // pixels per point
	zoom=min(zoom,vpoints/(v_bottom-v_top))
	zoom*=3
	
	variable pmin, pmax, qmin, qmax
	pmin=ceil(pixelX-zoom)
	pmax=floor(pixelX+zoom)
	qmin=ceil(pixelY-zoom)
	qmax=floor(pixelY+zoom)
	pmin=max(pmin, 0); qmin=max(qmin,0)
	pmax=min(pmax, (DimSize(w_img,0)-1))
	qmax=min(qmax, (DimSize(w_img,1)-1))
	variable i,j
	for (i=pmin;i<=pmax;i+=1)
		for (j=qmin;j<=qmax;j+=1)
			if (DimSize(w_img,2))
				w_img[i][j][0,2]=w[r]
			else
				w_img[i][j]=w[0]
			endif
		endfor
	endfor
	return 1
end

// hook function to deal with image editing
static function hookEditImg(STRUCT WMWinHookStruct &s)
	
	STRUCT PanelStatusStructure g
	fillPanelStructure(g)

	int key
	// doesn't work when mouse button is down?
	key = GetKeyState(0)
	if (key&1)
		s.doSetCursor = 1
		s.cursorCode = 20
	elseif (key&2)
		s.doSetCursor = 1
		s.cursorCode = 19
	endif
		
	int insert = (s.eventMod == 9) - (s.eventMod==5) // 1 for trace, -1 for bg
	if (! (s.eventCode == 3 && insert) ) // mousedown
		return 0
	endif

	DFREF dfr = root:Packages:Tracer
	wave w_img=GetImageRef()
	variable v_X, v_Y, p_X, p_Y
	
	STRUCT RGBcolor RGB
	if (insert == 1) // set image pixel to trace colour
		RGB = g.trace.rgb
	else // set image pixel to bg colour
		RGB = g.bg.rgb
	endif
		
	int keybit = insert == 1 ? 1 : 2
	int mousebutton
	do
		GetMouse /W=$s.WinName
		mousebutton = V_flag&1
		key = GetKeyState(0)&keybit
		
		if (!(mousebutton && key))
			return 0
		endif
		
		v_X=AxisValFromPixel("TracerGraph", "bottom", v_left )
		v_Y=AxisValFromPixel("TracerGraph", "left", v_top )
			
		// figure out which image pixel we're working on
		if(g.logY.value)
			wave TracerLogY=dfr:TracerLogY
			p_Y=BinarySearch(TracerLogY, v_Y)
		else
			p_Y=scaleToIndex(w_img, v_Y, 1)
		endif

		if(g.logX.value)
			wave TracerLogX=dfr:TracerLogX
			p_X=BinarySearch(TracerLogX, v_X)
		else
			p_X=scaleToIndex(w_img, v_X, 0)
		endif
		writePixelRGB(RGB, w_img, p_X, p_Y, g)
		DoUpdate
	while(1)
	return 0
end

// hook function to deal with cursors
static function hookTracer(STRUCT WMWinHookStruct &s)
		
	if (s.eventCode == 22 && s.eventMod&8) // mousewheel/touchpad + control/command key
		Make /D/free/N=3 wH, wV // free waves to hold axis minimum, maximum, and axis value for mouse location
		GetAxis /W=$s.WinName/Q bottom
		if (v_flag)
			return 0
		endif
		int logH = NumberByKey("log(x)", AxisInfo(s.WinName, "bottom"),"=")
		wH = {v_min, v_max, AxisValFromPixel(s.WinName, "bottom", s.mouseLoc.h)}
		wH = logH ? log(wH) : wH
		GetAxis /W=$s.WinName/Q left
		if (v_flag)
			return 0
		endif
		wV = {v_min, v_max, AxisValFromPixel(s.WinName, "left", s.mouseLoc.v)}
		int logV = NumberByKey("log(x)", AxisInfo(s.WinName, "left"),"=")
		wV = logV ? log(wV) : wV
		// alt/option key is accelerator
		variable expansion = s.eventMod&4 ? 1 - s.wheelDy / 5 : 1 - s.wheelDy / 50
		wH[0,1] = wH[2] - (wH[2] - wH[p]) * expansion
		wV[0,1] = wV[2] - (wV[2] - wV[p]) * expansion
		wH = logH ? alog(wH) : WH
		wV = logV ? alog(wV) : wV
		if (wH[1] > wH[0])
			SetAxis /W=$s.WinName bottom, wH[0], wH[1]
		else
			SetAxis /R/W=$s.WinName bottom, wH[0], wH[1]
		endif
		if (wV[1] > wV[0])
			SetAxis /W=$s.WinName left, wV[0], wV[1]
		else
			SetAxis /R/W=$s.WinName left, wV[0], wV[1]
		endif
		return 0
	endif
	
	if (s.eventCode!=7) // cursor moved
		return 0
	endif
	wave /Z w_img=GetImageRef()
	if (WaveExists(w_img)==0)
		return 0
	endif
		
	STRUCT RGBcolor RGB
	strswitch(s.cursorName)
		case "A":
			if ( WaveRefsEqual(w_img,CsrWaveRef(A,"TracerGraph")) )
				// update cursor colour
				getPixelRGB(RGB, w_img, pcsr(A,"TracerGraph"), qcsr(A,"TracerGraph") )
				contrastingColor(RGB)
				Cursor /M/C=(RGB.red,RGB.green,RGB.blue) /W=TracerGraph A
			endif
			return 0
		case "B":
			if ( WaveRefsEqual(w_img,CsrWaveRef(B,"TracerGraph")) )
				getPixelRGB(RGB, w_img, pcsr(B,"TracerGraph"), qcsr(B,"TracerGraph") )
				contrastingColor(RGB)
				Cursor /M/C=(RGB.red,RGB.green,RGB.blue) /W=TracerGraph B
			endif
			return 0
	endswitch
	return 0
end

static function contrastingColor(STRUCT RGBcolor &RGB)
	STRUCT HSLcolor HSL
	RGB2HSL(RGB, HSL)
	if(HSL.S<0.3)
		RGB.red=65535;RGB.green=0;RGB.blue=0
		return 1
	endif
	variable hue = round(6*HSL.H)
	hue -= 6*(hue>5)
	switch (hue)
		case 0: // red
			RGB.red=0;RGB.green=65535;RGB.blue=65535 // cyan
			break
		case 1: // yellow
			RGB.red=0;RGB.green=0;RGB.blue=65535 // blue
			break
		case 2: // green
			RGB.red=65535;RGB.green=0;RGB.blue=65535 // magenta
			break
		case 3: // cyan
			RGB.red=65535;RGB.green=0;RGB.blue=0 // red
			break
		case 4: // blue
			RGB.red=65535;RGB.green=65535;RGB.blue=0 // yellow
			break
		case 5: // magenta
			RGB.red=0;RGB.green=65535;RGB.blue=0 // lime
			break
	endswitch
	return 1
end

// see easyrgb.com
static function RGB2HSL(STRUCT RGBcolor &rgb, STRUCT HSLcolor &hsl)
	// R, G and B input range = 0 - 65535
	// H, S and L output range = 0 - 1
	variable red=rgb.red/65535
	variable green=rgb.green/65535
	variable blue=rgb.blue/65535

	variable var_Min = min(red,green,blue)
	variable var_Max = max(red,green,blue)
	variable del_Max = var_Max - var_Min
	
	variable H, S, L
	L = (var_Max + var_Min)/ 2
	if (del_Max == 0) // grey
	    H = 0
	    S = 0
	else // Chromatic data
		S = (L < 0.5) ? del_Max/(var_Max + var_Min) : del_Max/(2 - var_Max - var_Min)
		
		variable del_R = ( (var_Max - red)/6 + del_Max/2 ) / del_Max
		variable del_G = ( (var_Max - green)/6 + del_Max/2 ) / del_Max
		variable del_B = ( (var_Max - blue)/6 + del_Max/2 ) / del_Max
		
		if(red == var_Max )
			H = del_B - del_G
		elseif (green == var_Max )
			H = (1/3) + del_R - del_B
		elseif (blue == var_Max )
			H = (2/3) + del_G - del_R
		endif
		H += (H < 0)
		H -= (H > 1)
	endif
	hsl.H=H
	hsl.S=S
	hsl.L=L
	return 1
end

static structure XYZcolor
	float X, Y, Z
endstructure

static structure LABcolor
	float L, A, B
endstructure

static structure HSLcolor
	float H, S, L
endstructure

static function RGB2XYZ(STRUCT RGBcolor &rgb, STRUCT XYZcolor &xyz)
	// X, Y and Z output refer to a D65/2° standard illuminant.
	Make /free w={rgb.red, rgb.green, rgb.blue}
	w /= 65535
	w = w > 0.04045 ? ((w + 0.055) / 1.055)^2.4 : w / 12.92
	w *= 100
	
	xyz.X = w[0] * 0.4124 + w[1] * 0.3576 + w[2] * 0.1805
	xyz.Y = w[0] * 0.2126 + w[1] * 0.7152 + w[2] * 0.0722
	xyz.Z = w[0] * 0.0193 + w[1] * 0.1192 + w[2] * 0.9505
end

static function XYZ2LAB(STRUCT XYZcolor &xyz, STRUCT LABcolor &lab)
	// reference values for D65/2° standard illuminant.
	Make /free w={xyz.X, xyz.Y, xyz.Z}, XYZref={95.047,100,108.883}
	w /= XYZref
	w = (w > 0.008856) ? w^(1/3) :  7.787 * w + (16 / 116)
	lab.L = (116 * w[1]) - 16
	lab.A = 500 * (w[0] - w[1])
	lab.B = 200 * (w[1] - w[2])
end

static function RGB2LAB(STRUCT RGBcolor &rgb, STRUCT LABcolor &lab)
	// X, Y and Z output refer to a D65/2° standard illuminant.
	Make /free w={rgb.red, rgb.green, rgb.blue}
	w /= 65535
	w = w > 0.04045 ? ((w + 0.055) / 1.055)^2.4 : w / 12.92
	w *= 100
	Make /free/N=3 xyz
	xyz[0] = w[0] * 0.4124 + w[1] * 0.3576 + w[2] * 0.1805
	xyz[1] = w[0] * 0.2126 + w[1] * 0.7152 + w[2] * 0.0722
	xyz[2] = w[0] * 0.0193 + w[1] * 0.1192 + w[2] * 0.9505
	Make /free XYZref={95.047,100,108.883}
	w = xyz / XYZref
	w = (w > 0.008856) ? w^(1/3) :  7.787 * w + (16 / 116)
	lab.L = (116 * w[1]) - 16
	lab.A = 500 * (w[0] - w[1])
	lab.B = 200 * (w[1] - w[2])
end

static function deltaE94(STRUCT LABcolor &lab1, STRUCT LABcolor &lab2)
	variable xC1 = sqrt(lab1.A^2 + lab1.B^2)
	variable xC2 = sqrt(lab2.A^2 + lab2.B^2)
	variable xDL = lab2.L - lab1.L
	variable xDC = xC2 - xC1
	variable xDE = sqrt( (lab1.L - lab2.L)^2 + (lab1.A - lab2.A)^2 + (lab1.B - lab2.B)^2 )
	
	variable xDH = ( xDE * xDE ) - ( xDL * xDL ) - ( xDC * xDC )
	xDH = xDH > 0 ? sqrt( xDH ) : 0
	
	variable xSC = 1 + ( 0.045 * xC1 )
	variable xSH = 1 + ( 0.015 * xC1 )
	xDL /= 1
	xDC /= 1 * xSC
	xDH /= 1 * xSH
	return sqrt(xDL^2 + xDC^2 + xDH^2)
end

static function /WAVE getImageRef()
	string ImageNameString=StringFromList (0, ImageNameList("TracerGraph", ";" ))
	wave /Z w_img=ImageNameToWaveRef("TracerGraph",ImageNameString)
	return w_img
end

static function SetImgScale(int start)
	// make sure we don't already have cursors C-F on graph
	SetWindow TracerGraph hook(setscaleGUI)=$""
	KillControl /W=TracerGraph SetVarC; KillControl /W=TracerGraph SetVarD
	KillControl /W=TracerGraph SetVarE; KillControl /W=TracerGraph SetVarF
	Cursor /K/W=TracerGraph C; Cursor /K/W=TracerGraph D
	Cursor /K/W=TracerGraph E; Cursor /K/W=TracerGraph F
	ControlInfo /W=TracerPanel stopScale
	if(v_flag == 1)
		Button stopScale, win=TracerPanel,title="Off",fColor=(65535,65533,65534), Rename=startScale
	endif
	if (start == 0)
		return 0
	endif
	
	string strImage=StringFromList(0, ImageNameList("TracerGraph", ";"))
	wave wImage=ImageNameToWaveRef("TracerGraph", strImage)
	if (WaveExists(wImage)==0)
		return 0
	endif
	
	variable hSize=DimSize(wImage, 0), vSize=DimSize(wImage,1)
	// vertical hairs for X cursors
	Cursor /N=1/S=2/I/H=2/C=(0,0,65535)/P/W=TracerGraph C $strImage 0.1*hSize, 0.2*vSize
	Cursor /N=1/S=2/I/H=2/C=(0,0,65535)/P/W=TracerGraph D $strImage 0.9*hSize, 0.2*vSize
	// horizontal hairs for Y cursors
	Cursor /N=1/S=2/I/H=3/C=(0,65535,0)/P/W=TracerGraph E $strImage 0.2*hSize, 0.9*vSize
	Cursor /N=1/S=2/I/H=3/C=(0,65535,0)/P/W=TracerGraph F $strImage 0.2*hSize, 0.1*vSize
	
	STRUCT Point pt
	Make /free/T csr={"C","D","E","F"}
	int i
	for(i=0;i<4;i+=1)
		SetVariable $"SetVar"+csr[i] win=TracerGraph, title="", value=_NUM: i<2 ? hcsr($csr[i]) : vcsr($csr[i])
		SetVariable $"SetVar"+csr[i] win=TracerGraph, limits={-Inf,Inf,0}, size={40,10}, fsize=14, Proc=Tracer#Rescale
		SetVariable $"SetVar"+csr[i] win=TracerGraph, valueColor=(0,65535*(i>1),65535*(i<2))
	endfor
	
	SetWindow TracerGraph hook(setscaleGUI)=Tracer#hookSetScale, hookevents=4
	Button startScale, win=TracerPanel,title="On",fColor=(57346,65535,49151), Rename=stopScale
	
	// enter hookSetScale function with resize event to reposition setvars
	STRUCT WMWinHookStruct s
	s.WinName="TracerGraph"
	s.eventcode=6
	hookSetScale(s)
end

static function hookSetScale(STRUCT WMWinHookStruct &s)
	switch (s.eventcode)
		case 6:
		case 7:
		case 8:
			if(s.eventcode==7) // cursormoved
				if(GrepString(s.cursorName,"[C-F]")==0)
					return 0
				endif
				int isX = GrepString(s.cursorName,"[CD]")
					
				// keep axis coordinates within bounds of axes
				variable ptX, ptY
				GetAxis /Q/W=$s.WinName bottom
				ptX=max(hcsr($s.cursorName,s.WinName), min(V_Min,V_Max))
				ptX=min(ptX, max(V_Min,V_Max))
				GetAxis /Q/W=$s.WinName left
				ptY=max(vcsr($s.cursorName,s.WinName), min(V_Min,V_Max))
				ptY=min(ptY, max(V_Min,V_Max))
				
				STRUCT Point pt
				pt.h=PixelFromAxisVal(s.WinName, "bottom", ptX)
				pt.v=PixelFromAxisVal(s.WinName, "left", ptY)
				
				variable val = isX ? hcsr($s.cursorName,s.WinName) : vcsr($s.cursorName,s.WinName)
				SetVariable $"SetVar"+s.cursorName win=$s.WinName, value=_NUM:val, pos={pt.h-20,pt.v-10}, disable=0
				break
			endif
	
			// reposition or disable setvars when window is resized
			s.eventCode=7 // prepare to reenter this function with cursormoved eventcode
			Make /free/T csr={"C","D","E","F"}
			int i
			for(i=0;i<4;i+=1)
				s.cursorName=csr[i]
				variable csrpos = i>1 ? vcsr($s.cursorName, s.WinName) : hcsr($s.cursorName, s.WinName)
				GetAxis /Q/W=$s.WinName $SelectString(i>1, "bottom", "left")
				if(csrpos>min(v_max,v_min) && csrpos<max(v_max,v_min))
					hookSetScale(s)
				else
					SetVariable $"setvar"+(s.cursorName) win=$s.WinName, disable=1
				endif
			endfor
	endswitch
	return 0
end

// this is fired by the setscale setvars
static function Rescale(STRUCT WMSetVariableAction &s)
	if (s.eventCode != 8)
		return 0
	endif
	string strImage = StringFromList(0, ImageNameList(s.win, ";"))
	wave wImage = ImageNameToWaveRef(s.win, strImage)
	int isX = GrepString((s.ctrlName), "[CD]"), autoscale = 1
	ControlInfo /W=TracerPanel $SelectString(isX, "checkLogY", "checkLogX")
	int logAxis = v_value
	if (logAxis && s.dval<=0)
		SetVariable $s.ctrlName win=TracerGraph, value=_NUM:1
	endif
	
	string strAxis, info, flags
	info = ImageInfo(s.win, strImage, 0)
	strAxis = StringByKey(SelectString(isX, "YAXIS", "XAXIS"),info)
	flags = StringByKey("SETAXISFLAGS", AxisInfo("", strAxis))
	variable indexMin, indexMax
	if(GrepString(flags, "/")==0)
		GetAxis /Q $strAxis
		indexMin = scaleToIndex(wImage, V_min, 1-isX)
		indexMax = scaleToIndex(wImage, V_max, 1-isX)
		autoscale = 0
	endif
	
	variable ValCE, ValDF, delta, offset, oldDelta
	ControlInfo /W=$s.win $SelectString(isX, "SetVarE", "SetVarC")
	ValCE = V_Value
	ControlInfo /W=$s.win $SelectString(isX, "SetVarF", "SetVarD")
	ValDF = V_Value
		
	if(logAxis)
		DFREF dfr = root:Packages:Tracer
		wave w = dfr:$SelectString(isX, "TracerLogY", "TracerLogX")
		variable p1 = isX ? pcsr(C) : qcsr(E)
		variable p2 = isX ? pcsr(D) : qcsr(F)
		string strNote=note(wImage)
		w=alog(log(ValCE)+(p-0.5-p1)*(log(ValDF)-log(ValCE))/(p2-p1))
		variable low=alog(log(ValCE)+(-1-p1)*(log(ValDF)-log(ValCE))/(p2-p1))
		variable high=alog(log(ValCE)+(DimSize(w,0)-1-p1)/(p2-p1)*(log(ValDF)-log(ValCE)))
		strNote=ReplaceNumberByKey(SelectString(isX,"YHIGH","XHIGH"), strNote, low)
		strNote=ReplaceNumberByKey(SelectString(isX,"YLOW","XLOW"), strNote, high)
		note /K wImage, strNote
		
		if (autoscale)
			if (w[1]>w[0])
				if (isX)
					SetAxis/A /W=TracerGraph bottom
				else
					SetAxis/A/R /W=TracerGraph left
				endif
			else
				if (isX)
					SetAxis/A/R /W=TracerGraph bottom
				else
					SetAxis/A /W=TracerGraph left
				endif
			endif
		endif
		return 0
	endif
			
	delta = isX ? (ValDF-ValCE)/(pcsr(D)-pcsr(C)) : (ValDF-ValCE)/(qcsr(F)-qcsr(E))
	offset = isX ? ValCE - delta*pcsr(C) : ValCE - delta*qcsr(E)
	oldDelta = DimDelta(wImage, 1-isX)
	if(isX)
		SetScale /P x, offset, delta , wImage
	else
		SetScale /P y, offset, delta , wImage
	endif
	if (autoscale==0)
		v_min = IndexToScale(wImage, indexMin, 1-isX)
		v_max = IndexToScale(wImage, indexMax, 1-isX)
		if(v_min > v_max)
			SetAxis /R/W=$s.win $strAxis v_min, v_max
		else
			SetAxis /W=$s.win $strAxis v_min, v_max
		endif
	endif
	if(sign(oldDelta) != sign(delta))
		// switch the axis limits so that image is not flipped
		if (GrepString(flags, "/R"))
			SetAxis /A/W=$s.win $strAxis
		elseif (GrepString(flags, "/A"))
			SetAxis /A/R/W=$s.win $strAxis
		endif
	endif
	return 0
end

// copy and paste image scale

menu "TracePopup", dynamic
	tracer#ScaleMenuString(0), /Q, tracer#CopyImageScale()
	tracer#ScaleMenuString(1), /Q, tracer#PasteImageScale()
end

static function /S ScaleMenuString(int paste)
	if (WinType("")!=1)
		return "" // don't do anything if Igor is just rebuilding the menu
	endif
	// figure out graph and trace names
	GetLastUserMenuInfo
	if(strlen(ImageNameList(s_graphname, ";")) == 0)
		return ""
	endif
	if (paste == 0)
		return "Copy Image Scale"
	elseif (cmpstr(GetScrapText()[0,7], "SetScale")==0)
		return "Paste Image Scale"
	endif
	return ""
end

static function CopyImageScale()
	GetLastUserMenuInfo
	wave /Z w = ImageNameToWaveRef(s_graphname, StringFromList(0, ImageNameList(s_graphname, ";")))
	if (WaveExists(w)==0)
		return 0
	endif
	string cmd = ""
	sprintf cmd, "SetScale /P x, %g, %g, ###; SetScale /P y, %g, %g, ###;", DimOffset(w, 0), DimDelta(w, 0), DimOffset(w, 1), DimDelta(w, 1)
	PutScrapText cmd
end

static function PasteImageScale()
	GetLastUserMenuInfo
	string strImage = StringFromList(0, ImageNameList(s_graphname, ";"))
	wave /Z wImage = ImageNameToWaveRef(s_graphname, strImage)
	if (WaveExists(wImage) == 0)
		return 0
	endif
	
	variable x0, dx, y0, dy
	sscanf GetScrapText(), "SetScale /P x, %g, %g, ###; SetScale /P y, %g, %g, ###;", x0, dx, y0, dy
	if (V_flag != 4)
		return 0
	endif
	
	int autoscale, dim
	string strAxis = "", info = "", flags = ""
	info = ImageInfo("", strImage, 0)
	variable indexMin, indexMax, delta, oldDelta
	
	for (dim=0;dim<2;dim+=1)
		autoscale = 1
		straxis = StringByKey(SelectString(dim, "XAXIS", "YAXIS"), info)
		flags = StringByKey("SETAXISFLAGS", AxisInfo("", strAxis))
		
		oldDelta = DimDelta(wImage, dim)
		delta = dim ? dY : dX
		
		if (GrepString(flags, "/")==0)
			GetAxis /Q $strAxis
			indexMin = scaleToIndex(wImage, V_min, dim)
			indexMax = scaleToIndex(wImage, V_max, dim)
			autoscale = 0
		endif
		if (dim == 0)
			SetScale /P x, x0, delta , wImage
		else
			SetScale /P y, y0, delta , wImage
		endif
		if (autoscale==0)
			v_min = IndexToScale(wImage, indexMin, dim)
			v_max = IndexToScale(wImage, indexMax, dim)
			if (v_min > v_max)
				SetAxis /R $strAxis v_min, v_max
			else
				SetAxis $strAxis v_min, v_max
			endif
		endif
		if (sign(oldDelta) != sign(delta))
			// switch the axis limits so that image is not flipped
			if (GrepString(flags, "/R"))
				SetAxis /A $strAxis
			elseif (GrepString(flags, "/A"))
				SetAxis /A/R $strAxis
			endif
		endif
	endfor
end
